import string
import warnings
from json import loads

from .exceptions import LexerError, EmptyExpressionError


class Lexer(object):
    START_IDENTIFIER = set(string.ascii_letters + "_")
    VALID_IDENTIFIER = set(string.ascii_letters + string.digits + "_")
    VALID_NUMBER = set(string.digits)
    WHITESPACE = set(" \t\n\r")
    SIMPLE_TOKENS = {
        ".": "dot",
        "*": "star",
        "]": "rbracket",
        ",": "comma",
        ":": "colon",
        "@": "current",
        "(": "lparen",
        ")": "rparen",
        "{": "lbrace",
        "}": "rbrace",
    }

    def tokenize(self, expression):
        self._initialize_for_expression(expression)
        while self._current is not None:
            if self._current in self.SIMPLE_TOKENS:
                yield {
                    "type": self.SIMPLE_TOKENS[self._current],
                    "value": self._current,
                    "start": self._position,
                    "end": self._position + 1,
                }
                self._next()
            elif self._current in self.START_IDENTIFIER:
                start = self._position
                buff = self._current
                while self._next() in self.VALID_IDENTIFIER:
                    buff += self._current
                yield {
                    "type": "unquoted_identifier",
                    "value": buff,
                    "start": start,
                    "end": start + len(buff),
                }
            elif self._current in self.WHITESPACE:
                self._next()
            elif self._current == "[":
                start = self._position
                next_char = self._next()
                if next_char == "]":
                    self._next()
                    yield {
                        "type": "flatten",
                        "value": "[]",
                        "start": start,
                        "end": start + 2,
                    }
                elif next_char == "?":
                    self._next()
                    yield {
                        "type": "filter",
                        "value": "[?",
                        "start": start,
                        "end": start + 2,
                    }
                else:
                    yield {
                        "type": "lbracket",
                        "value": "[",
                        "start": start,
                        "end": start + 1,
                    }
            elif self._current == "'":
                yield self._consume_raw_string_literal()
            elif self._current == "|":
                yield self._match_or_else("|", "or", "pipe")
            elif self._current == "&":
                yield self._match_or_else("&", "and", "expref")
            elif self._current == "`":
                yield self._consume_literal()
            elif self._current in self.VALID_NUMBER:
                start = self._position
                buff = self._consume_number()
                yield {
                    "type": "number",
                    "value": int(buff),
                    "start": start,
                    "end": start + len(buff),
                }
            elif self._current == "-":
                # Negative number.
                start = self._position
                buff = self._consume_number()
                if len(buff) > 1:
                    yield {
                        "type": "number",
                        "value": int(buff),
                        "start": start,
                        "end": start + len(buff),
                    }
                else:
                    raise LexerError(
                        lexer_position=start,
                        lexer_value=buff,
                        message="Unknown token '%s'" % buff,
                    )
            elif self._current == '"':
                yield self._consume_quoted_identifier()
            elif self._current == "<":
                yield self._match_or_else("=", "lte", "lt")
            elif self._current == ">":
                yield self._match_or_else("=", "gte", "gt")
            elif self._current == "!":
                yield self._match_or_else("=", "ne", "not")
            elif self._current == "=":
                if self._next() == "=":
                    yield {
                        "type": "eq",
                        "value": "==",
                        "start": self._position - 1,
                        "end": self._position,
                    }
                    self._next()
                else:
                    if self._current is None:
                        # If we're at the EOF, we never advanced
                        # the position so we don't need to rewind
                        # it back one location.
                        position = self._position
                    else:
                        position = self._position - 1
                    raise LexerError(
                        lexer_position=position,
                        lexer_value="=",
                        message="Unknown token '='",
                    )
            else:
                raise LexerError(
                    lexer_position=self._position,
                    lexer_value=self._current,
                    message="Unknown token %s" % self._current,
                )
        yield {"type": "eof", "value": "", "start": self._length, "end": self._length}

    def _consume_number(self):
        start = self._position
        buff = self._current
        while self._next() in self.VALID_NUMBER:
            buff += self._current
        return buff

    def _initialize_for_expression(self, expression):
        if not expression:
            raise EmptyExpressionError()
        self._position = 0
        self._expression = expression
        self._chars = list(self._expression)
        self._current = self._chars[self._position]
        self._length = len(self._expression)

    def _next(self):
        if self._position == self._length - 1:
            self._current = None
        else:
            self._position += 1
            self._current = self._chars[self._position]
        return self._current

    def _consume_until(self, delimiter):
        # Consume until the delimiter is reached,
        # allowing for the delimiter to be escaped with "\".
        start = self._position
        buff = ""
        self._next()
        while self._current != delimiter:
            if self._current == "\\":
                buff += "\\"
                self._next()
            if self._current is None:
                # We're at the EOF.
                raise LexerError(
                    lexer_position=start,
                    lexer_value=self._expression[start:],
                    message="Unclosed %s delimiter" % delimiter,
                )
            buff += self._current
            self._next()
        # Skip the closing delimiter.
        self._next()
        return buff

    def _consume_literal(self):
        start = self._position
        lexeme = self._consume_until("`").replace("\\`", "`")
        try:
            # Assume it is valid JSON and attempt to parse.
            parsed_json = loads(lexeme)
        except ValueError:
            try:
                # Invalid JSON values should be converted to quoted
                # JSON strings during the JEP-12 deprecation period.
                parsed_json = loads('"%s"' % lexeme.lstrip())
                warnings.warn(
                    "deprecated string literal syntax", PendingDeprecationWarning
                )
            except ValueError:
                raise LexerError(
                    lexer_position=start,
                    lexer_value=self._expression[start:],
                    message="Bad token %s" % lexeme,
                )
        token_len = self._position - start
        return {
            "type": "literal",
            "value": parsed_json,
            "start": start,
            "end": token_len,
        }

    def _consume_quoted_identifier(self):
        start = self._position
        lexeme = '"' + self._consume_until('"') + '"'
        try:
            token_len = self._position - start
            return {
                "type": "quoted_identifier",
                "value": loads(lexeme),
                "start": start,
                "end": token_len,
            }
        except ValueError as e:
            error_message = str(e).split(":")[0]
            raise LexerError(
                lexer_position=start, lexer_value=lexeme, message=error_message
            )

    def _consume_raw_string_literal(self):
        start = self._position
        lexeme = self._consume_until("'").replace("\\'", "'")
        token_len = self._position - start
        return {"type": "literal", "value": lexeme, "start": start, "end": token_len}

    def _match_or_else(self, expected, match_type, else_type):
        start = self._position
        current = self._current
        next_char = self._next()
        if next_char == expected:
            self._next()
            return {
                "type": match_type,
                "value": current + next_char,
                "start": start,
                "end": start + 1,
            }
        return {"type": else_type, "value": current, "start": start, "end": start}



from __future__ import annotations


def compact_try(expr_fn, fallback_fn=None):
    try:
        return expr_fn()
    except Exception as exc:
        if fallback_fn is None:
            fallback_member = getattr(exc, "__fallback__", None)
            if callable(fallback_member):
                return fallback_member()
            return exc
        return fallback_fn(exc)



from __future__ import annotations

import subprocess


class SubprocessCapture:
    def __init__(self, cmd: str) -> None:
        self.cmd = cmd

    def __pipeline__(self, input_data):
        try:
            if input_data is None:
                completed = subprocess.run(
                    self.cmd,
                    shell=True,
                    check=True,
                    text=True,
                    stdout=subprocess.PIPE,
                )
            else:
                if not isinstance(input_data, (str, bytes)):
                    input_data = str(input_data)
                completed = subprocess.run(
                    self.cmd,
                    shell=True,
                    check=True,
                    text=True,
                    input=input_data,
                    stdout=subprocess.PIPE,
                )
            return completed.stdout.rstrip("\n")
        except subprocess.CalledProcessError as exc:

            def __fallback(exc=exc):
                raise exc

            exc.__fallback__ = __fallback
            raise


class SubprocessStatus:
    def __init__(self, cmd: str) -> None:
        self.cmd = cmd

    def __pipeline__(self, input_data):
        try:
            if input_data is None:
                subprocess.run(self.cmd, shell=True, check=True)
            else:
                if not isinstance(input_data, (str, bytes)):
                    input_data = str(input_data)
                subprocess.run(
                    self.cmd,
                    shell=True,
                    check=True,
                    text=True,
                    input=input_data,
                )
            return 0
        except subprocess.CalledProcessError as exc:

            def __fallback(exc=exc):
                return exc.returncode

            exc.__fallback__ = __fallback
            raise



from __future__ import annotations

import json as _json
import sys as _sys

from ..vendor import jmespath


class StructuredAccessor:
    def __init__(self, query: str) -> None:
        self.query = query

    def __pipeline__(self, obj):
        if not hasattr(obj, "__structured__"):
            raise TypeError(
                "Pipeline target must implement __structured__, "
                f"got {type(obj).__name__}"
            )
        return obj.__structured__(self.query)


class JsonObject:
    def __init__(self, data) -> None:
        self.data = data

    def __structured__(self, query: str):
        return jmespath.search(query, self.data)

    def __repr__(self) -> str:
        return _json.dumps(self.data, indent=2)


class JsonPipelineWrapper:
    """Wrapper for json() to support pipeline operator without blocking stdin."""

    def __pipeline__(self, input_data):
        return json(input_data)

    def __structured__(self, query: str):
        data = json(_sys.stdin)
        return data.__structured__(query)

    def __repr__(self) -> str:
        data = json(_sys.stdin)
        return repr(data)


def json(input_data=None):
    """Parse JSON from various input sources."""
    if input_data is None:
        return JsonPipelineWrapper()

    if isinstance(input_data, str):
        try:
            data = _json.loads(input_data)
        except _json.JSONDecodeError:
            with open(input_data, "r", encoding="utf-8") as handle:
                data = _json.load(handle)
    elif hasattr(input_data, "read"):
        content = input_data.read()
        if isinstance(content, bytes):
            content = content.decode("utf-8")
        data = _json.loads(content)
    elif isinstance(input_data, (dict, list, int, float, bool)) or input_data is None:
        data = input_data
    else:
        raise TypeError(
            f"json() input must be JSON-compatible, got {type(input_data).__name__}"
        )

    return JsonObject(data)



from ._native import compile, exec, parse

try:
    from importlib.metadata import version

    __version__ = version("snail-lang")
except Exception:  # pragma: no cover - during development
    __version__ = "0.0.0"

__all__ = ["compile", "exec", "parse", "__version__"]



from __future__ import annotations

from .compact_try import compact_try
from .regex import regex_compile, regex_search
from .structured_accessor import (
    JsonObject,
    JsonPipelineWrapper,
    StructuredAccessor,
    json,
)
from .subprocess import SubprocessCapture, SubprocessStatus

__all__ = ["install_helpers"]


def install_helpers(globals_dict: dict) -> None:
    globals_dict["__snail_compact_try"] = compact_try
    globals_dict["__snail_regex_search"] = regex_search
    globals_dict["__snail_regex_compile"] = regex_compile
    globals_dict["__SnailSubprocessCapture"] = SubprocessCapture
    globals_dict["__SnailSubprocessStatus"] = SubprocessStatus
    globals_dict["__SnailStructuredAccessor"] = StructuredAccessor
    globals_dict["__SnailJsonObject"] = JsonObject
    globals_dict["__SnailJsonPipelineWrapper"] = JsonPipelineWrapper
    globals_dict["json"] = json



from __future__ import annotations

import argparse
import sys
from pathlib import Path

from . import __version__, exec, parse


def _build_parser() -> argparse.ArgumentParser:
    return argparse.ArgumentParser(
        prog="snail",
        description="Snail programming language interpreter",
        usage="snail [options] -f <file> [args]...\n       snail [options] <code> [args]...",
        add_help=True,
    )


def main(argv: list[str] | None = None) -> int:
    parser = _build_parser()
    parser.add_argument("-f", dest="file", metavar="file")
    parser.add_argument("-a", "--awk", action="store_true")
    parser.add_argument("-P", "--no-print", action="store_true")
    parser.add_argument("--parse-only", action="store_true")
    parser.add_argument("-v", "--version", action="store_true")
    parser.add_argument("args", nargs=argparse.REMAINDER)

    namespace = parser.parse_args(argv)

    if namespace.version:
        print(__version__)
        return 0

    mode = "awk" if namespace.awk else "snail"

    if namespace.file:
        path = Path(namespace.file)
        try:
            source = path.read_text()
        except OSError as exc:
            print(f"failed to read {path}: {exc}", file=sys.stderr)
            return 1
        filename = str(path)
        args = [filename, *namespace.args]
    else:
        if not namespace.args:
            print("no input provided", file=sys.stderr)
            return 1
        source = namespace.args[0]
        filename = "<cmd>"
        args = ["--", *namespace.args[1:]]

    if namespace.parse_only:
        parse(source, mode=mode, filename=filename)
        return 0

    return exec(
        source,
        argv=args,
        mode=mode,
        auto_print=not namespace.no_print,
        filename=filename,
    )


if __name__ == "__main__":
    raise SystemExit(main())
